/*
 * Copyright (C) 2007 Rufus Garton Smith
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/** Task Manager Module
 * Features:
 * - View a list of processes and associated runtime information
 * - Kill processes
 * - Renice processes
 * 
 * "Features":
 * - Leaks quite a lot of memory,
 *   TWindows aren't being freed when their parent menus are removed,
 *   which means there's a whole hierachy of things not being freed.
 *   If I force the Window to be freed, it causes terrible memory corruption, and I don't know why.
 * - Crashes under certain conditions, such as when looking at the details of a process as it terminates
 *   This appears to be related to the above issue
 */

#include <ctype.h>
#include <dirent.h>
#include <errno.h>
#include <fcntl.h>
#include <limits.h>
#include <signal.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/resource.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>

#include <pz.h>

#define TM_MENU_ROOT "/Extras/Utilities/Task Manager"

#define TM_VIEW_MODE 1
#define TM_UPDATE_SPEED 2

#define TM_VM_NORMAL 0
#define TM_VM_SPLIT 1

// Define this symbol to prevent the module from actually killing/renicing anything
//#define TASKMAN_TESTING

typedef struct __proc_ent {
	struct __proc_ent *next;
	ttk_menu_item *item;
	pid_t pid;
	char *comm;
	char state;
	unsigned long utime;
	unsigned long ktime;
	int cent_stamp;
	long nice;
	unsigned long vsize;
	unsigned int uload;
	unsigned int kload;
} PROC_ENT;

typedef struct __updater_data {
	TWindow *wwid;
	PROC_ENT *before;
	PROC_ENT *after;
} UPDATER_DATA;

typedef struct __renice_data {
	int new_nice;
	int old_nice;
	PROC_ENT *ent;
} RENICE_DATA;

static PzModule *tm_module;
//static PzConfig *tm_config;

static void tm_init();
static void tm_cleanup();

static PzWindow *tm_root_mh(ttk_menu_item *item);
static int tm_root_updater_frame(TWidget *wid);
static void tm_root_updater_rment(PROC_ENT *cur, UPDATER_DATA *data);
static void tm_root_updater_swapentdata(PROC_ENT *a, PROC_ENT *b);
static void *tm_root_updater_md(TWindow *wwid);

static int tm_update_proc_ent(PROC_ENT *ent);
static void tm_free_proc_ent(PROC_ENT *ent);
static PROC_ENT *tm_new_proc_ent(PROC_ENT *next);

static PzWindow *tm_proc_mh(ttk_menu_item *item);

static PzWindow *tm_renice_mh(ttk_menu_item *item);
static int tm_renice_slider_down(TWidget *wid, int button);
static TWidget *tm_new_renice_monitor_widget(PROC_ENT *ent, int init_prio);
static void tm_renice_monitor_destroy(TWidget *wid);
static void tm_renice_monitor_draw(TWidget *wid, ttk_surface srf);
static void tm_renice_monitor_do_renice(TWidget *wid);
static void tm_renice_widget_callback(int cdata, int val);

static PzWindow *tm_signal_mh(ttk_menu_item *item);

static PzWindow *tm_info_mh(ttk_menu_item *item);
static TWidget *tm_new_info_widget(PROC_ENT *ent);
static int tm_info_down(TWidget *wid, int button);
static int tm_info_frame(TWidget *wid);
static void tm_info_draw(TWidget *wid, ttk_surface srf);

static PzWindow *tm_kill_mh(ttk_menu_item *item);

static int tm_selected_updater_frame(TWidget *wid);


#define KILL_ITEM(signal,common) { #signal, { tm_kill_mh }, .group_name = (common?"Common":"Other"), .flags = TTK_MENU_ICON_EXE, .cdata=signal }
static const ttk_menu_item tm_kill_items[29] =
{
	KILL_ITEM(SIGTERM, 1),
	KILL_ITEM(SIGKILL, 1),
	KILL_ITEM(SIGINT, 1),
	KILL_ITEM(SIGHUP, 1),
	KILL_ITEM(SIGSTOP, 1),
	KILL_ITEM(SIGCONT, 1),
	KILL_ITEM(SIGABRT, 0),
	KILL_ITEM(SIGALRM, 0),
	KILL_ITEM(SIGFPE, 0),
	KILL_ITEM(SIGILL, 0),
	KILL_ITEM(SIGPIPE, 0),
	KILL_ITEM(SIGQUIT, 0),
	KILL_ITEM(SIGSEGV, 0),
	KILL_ITEM(SIGCHLD, 0),
	KILL_ITEM(SIGTSTP, 0),
	KILL_ITEM(SIGTTIN, 0),
	KILL_ITEM(SIGTTOU, 0),
	KILL_ITEM(SIGBUS, 0),
	KILL_ITEM(SIGPOLL, 0),
	KILL_ITEM(SIGPROF, 0),
	KILL_ITEM(SIGSYS, 0),
	KILL_ITEM(SIGTRAP, 0),
	KILL_ITEM(SIGURG, 0),
	KILL_ITEM(SIGUSR1, 0),
	KILL_ITEM(SIGUSR2, 0),
	KILL_ITEM(SIGVTALRM, 0),
	KILL_ITEM(SIGXCPU, 0),
	KILL_ITEM(SIGXFSZ, 0),
	{0}
};

static const ttk_menu_item tm_proc_items[4] = {
	{"Information", {tm_info_mh}, TTK_MENU_ICON_SUB },
	{"Kill", {tm_signal_mh}, TTK_MENU_ICON_SUB },
	{"Renice", {tm_renice_mh}, TTK_MENU_ICON_SUB },
	{0}
};


static int tm_selected_updater_frame(TWidget *wid) {
	PROC_ENT *selected = ((ttk_menu_item*)wid->data)->data;
	char path[40];
	snprintf(path, 40, "/proc/%d", selected->pid);
	if(access(path, F_OK)) {
		// FIXME: freeing the window causes instability
		// LEAK: The menu should be freeing the data here, but isn't,
		//ttk_menu_item *item = wid->data;
		//item->sub = NULL;
		//item->flags &= ~TTK_MENU_MADESUB;
		ttk_hide_window(wid->win);
		//ttk_free_window(wid->win);
	}
	return 0;
}

static int tm_info_down(TWidget *wid, int button) {
	switch(button) {
	case TTK_BUTTON_MENU:
		ttk_hide_window(wid->win);
		return TTK_EV_CLICK;
	case TTK_BUTTON_ACTION:
		ttk_widget_set_inv_fps(wid, 2);
		wid->frame(wid);
		return TTK_EV_CLICK;
	}
	return TTK_EV_UNUSED;
}

static void tm_info_draw(TWidget *wid, ttk_surface srf) {
	PROC_ENT *data = wid->data;
	ttk_color black = ttk_makecol(BLACK);
	char text[40];
	char *state;
	int y = 0;
	snprintf(text, sizeof(text), "%s: %s", _("Name"), data->comm);
	ttk_text(srf, ttk_textfont, 0, y, black, text);
	y+= ttk_text_height(ttk_textfont)+1;
	switch(data->state) {
		case 'R':
			state = _("Running");
			break;
		default:
		case 'S':
			state = _("Sleeping");
			break;
		case 'D':
			state = _("Disk Sleep");
			break;
		case 'Z':
			state = _("Zombie");
			break;
		case 'T':
			state = _("Stopped");
			break;
		case 'W':
			state = _("Paging");
	}
	snprintf(text, sizeof(text), "%s: %s", _("State"), state);
	ttk_text(srf, ttk_textfont, 0, y, black, text);
	y+= ttk_text_height(ttk_textfont)+1;
	
	int load = (data->uload * 10000) >> 16;
	int load_frac = load % 100;
	load /= 100;
	snprintf(text, sizeof(text), "%s: %d.%.2d%%", _("User"), load, load_frac);
	ttk_text(srf, ttk_textfont, 0, y, black, text);
	y+= ttk_text_height(ttk_textfont)+1;
	load = (data->kload * 10000) >> 16;
	load_frac = load % 100;
	load /= 100;
	snprintf(text, sizeof(text), "%s: %d.%.2d%%", _("Kernel"), load, load_frac);
	ttk_text(srf, ttk_textfont, 0, y, black, text);
	y+= ttk_text_height(ttk_textfont)+1;
	snprintf(text, sizeof(text), "%s: %ld", _("Nice"), data->nice);
	ttk_text(srf, ttk_textfont, 0, y, black, text);
	y+= ttk_text_height(ttk_textfont)+1;
	snprintf(text, sizeof(text), "%s: %lukb", _("VM Size"), (data->vsize) >> 10);
	ttk_text(srf, ttk_textfont, 0, y, black, text);
}

static int tm_info_frame(TWidget *wid) {
	PROC_ENT *data = wid->data;
	switch(tm_update_proc_ent(data)) {
	case 0:
		wid->dirty++;
		break;
	case 1:
	case -1:
		ttk_hide_window(wid->win);
		break;
	}
	return 0;
}

static TWidget *tm_new_info_widget(PROC_ENT *ent) {
	TWidget *ret = ttk_new_widget(0, 0);
	if(!ret) goto E0;
	ret->w = ttk_text_width(ttk_textfont, _("Name: WWWWWWWWWWWW"));
//	ret->w = ttk_text_width(ttk_textfont, _("State: Sleeping (Uninterruptable)"));
	ret->h = ttk_text_height(ttk_textfont)*6+5;
	ret->data = ent;
	ret->draw = tm_info_draw;
	ret->frame = tm_info_frame;
	ret->down = tm_info_down;
	ret->focusable = 1;
	ttk_widget_set_inv_fps(ret, 2);
	return ret;

E0:	return NULL;
}

static PzWindow *tm_info_mh(ttk_menu_item *item) {
	PROC_ENT *selected = item->data;
	char title[40];
	snprintf(title, sizeof(title), "%s (%s)", _("Information"), selected->comm);
	PzWindow *ret = pz_new_window(title, PZ_WINDOW_POPUP);
	TWidget *wid = tm_new_info_widget(selected);
	if(!wid) goto E0;
	ttk_add_widget(ret, wid);
	return pz_finish_window(ret);

E0:	return TTK_MENU_DONOTHING;
}

static PzWindow *tm_kill_mh(ttk_menu_item *item) {
	PROC_ENT *selected = item->data;
#ifndef TASKMAN_TESTING
 	if(kill(selected->pid, item->cdata) && errno == EPERM)
 		pz_error(_("Insufficient privileges to kill that process!"));
	// All other error conditions are inconsequential
#else
	printf("kill -%s %d\n", item->name, selected->pid);
#endif
	return TTK_MENU_UPONE;
}

static PzWindow *tm_signal_mh(ttk_menu_item *item) {
	PROC_ENT *selected = item->data;
	char title[40];
	snprintf(title, sizeof(title), "%s (%s)", _("Kill"), selected->comm);
	PzWindow *ret = pz_new_window(title, PZ_WINDOW_POPUP);
	if(!ret) { pz_error(_("Unable to create TWindow!")); goto E0; }
	ttk_menu_item *items = malloc(sizeof(tm_kill_items));
	memcpy(items, tm_kill_items, sizeof(tm_kill_items));
	int i;
	for(i=0; items[i].name; i++)
		items[i].data = selected;
	TWidget *wid = ttk_new_menu_widget(items, ttk_menufont, ret->w, ret->h);
	if(!wid) { pz_error(_("Unable to create TWidget!")); goto E1; }
	ttk_menu_sort_groups(wid);
	ttk_add_widget(ret, wid);
	
	// Add invisible widget for timer events
	wid = ttk_new_widget(0,0);
	if(!wid) { pz_error(_("Unable to create TWidget!")); goto E1; }
	wid->data = item;
	wid->frame = tm_selected_updater_frame;
	ttk_add_widget(ret, wid);
	ttk_widget_set_inv_fps(wid, 2);
	return pz_finish_window(ret);

E1:	ttk_free_window(ret);
E0:	return TTK_MENU_DONOTHING;
}

static void tm_renice_widget_callback(int cdata, int val) {
	// HACK: Need a better way to pass a ptr without using a global, but don't want to have to build a map of indices to ptrs
	TWidget *wid = (TWidget*)cdata;
	RENICE_DATA *data = wid->data;
	wid->dirty++;
	data->new_nice = val;
}

static void tm_renice_monitor_do_renice(TWidget *wid) {
	RENICE_DATA *data = wid->data;
#ifndef TASKMAN_TESTING
	if(data->new_nice == data->old_nice)
		return;
	if(setpriority(PRIO_PROCESS, data->ent->pid, data->new_nice) && (errno == EACCES || errno == EPERM))
		pz_error(errno == EACCES ? _("Insufficient privileges to reduce the nice value of a process") :
				_("Insufficient privileges to change nice value of that process"));
#else
	printf("renice %d -p %d\n", data->new_nice, data->ent->pid);
#endif
}

static void tm_renice_monitor_draw(TWidget *wid, ttk_surface srf) {
	RENICE_DATA *data = wid->data;
	ttk_color black = ttk_makecol(BLACK);
	char text[40];
	snprintf(text, sizeof(text), "%s: %d", _("Old nice value"), data->old_nice);
	ttk_text(srf, ttk_textfont, 0, 0, black, text);
	snprintf(text, sizeof(text), "%s: %d", _("New nice value"), data->new_nice);
	ttk_text(srf, ttk_textfont, 0, ttk_text_height(ttk_textfont)+1, black, text);
}

static void tm_renice_monitor_destroy(TWidget *wid) {
	free(wid->data);
}

static TWidget *tm_new_renice_monitor_widget(PROC_ENT *ent, int init_prio) {
	TWidget *ret = ttk_new_widget(0, 0);
	if(!ret) goto E0;
	ret->w = ttk_text_width(ttk_textfont, _("New nice value: -20"));
	ret->h = ttk_text_height(ttk_textfont)*2+1;
	RENICE_DATA *data = malloc(sizeof(RENICE_DATA));
	if(!data) goto E1;
	data->new_nice = data->old_nice = init_prio;
	data->ent = ent;
	ret->data = data;
	ret->draw = tm_renice_monitor_draw;
	ret->destroy = tm_renice_monitor_destroy;
	return ret;

E1:	free(ret);
E0:	return NULL;
}

static int tm_renice_slider_down(TWidget *wid, int button) {
	if(button == TTK_BUTTON_ACTION) {
		tm_renice_monitor_do_renice(wid->win->widgets->next->v);
		ttk_hide_window(wid->win);
		return TTK_EV_CLICK;
	} else {
		return ttk_slider_down(wid, button);
	}
}

static PzWindow *tm_renice_mh(ttk_menu_item *item) {
	// TODO: the renice window should close if the process dies
	PROC_ENT *selected = item->data;
	char title[40];
	snprintf(title, sizeof(title), "%s (%s)", _("Renice"), selected->comm);
	PzWindow *ret = pz_new_window(title, PZ_WINDOW_POPUP);
	if(!ret) { pz_error(_("Unable to create TWindow!")); goto E0; }
	// Since the nice value may change, we should recreate the window every time so the widgets can reflect those changes
	ret->data = 0x12345678;
	int nzero = sysconf(_SC_NZERO);
	int priority = getpriority(PRIO_PROCESS, selected->pid);
	TWidget *monitor = tm_new_renice_monitor_widget(selected, priority);
	if(!monitor) { pz_error(_("Unable to create TWidget!")); goto E1; }
	TWidget *slider = ttk_new_slider_widget(0, monitor->h + 2, ret->w * 2 / 3, -nzero, nzero - 1, &((RENICE_DATA*)monitor->data)->new_nice, NULL);
	if(!slider) { pz_error(_("Unable to create TWidget!")); goto E2; }
	ttk_slider_set_callback(slider, tm_renice_widget_callback, (int)monitor);
	slider->down = tm_renice_slider_down;
	
	
	ttk_add_widget(ret, slider);
	ttk_add_widget(ret, monitor);
		
	// Add invisible widget for timer events
	TWidget *wid = ttk_new_widget(0,0);
	if(!wid) { pz_error(_("Unable to create TWidget!")); goto E1; }
	wid->data = item;
	wid->frame = tm_selected_updater_frame;
	ttk_add_widget(ret, wid);
	ttk_widget_set_inv_fps(wid, 2);
	return pz_finish_window(ret);

E2:	ttk_free_widget(monitor);
E1:	ttk_free_window(ret);
E0:	return TTK_MENU_DONOTHING;
}

static PzWindow *tm_proc_mh(ttk_menu_item *item) {
	PROC_ENT *selected = item->data;
	char title[40];
	snprintf(title, sizeof(title), "%s (%s)", _("Process"), selected->comm);
	PzWindow *ret = pz_new_window(title, PZ_WINDOW_NORMAL);
	if(!ret) { pz_error(_("Unable to create TWindow!")); goto E0; }
	ttk_menu_item *items = malloc(sizeof(tm_proc_items));
	memcpy(items, tm_proc_items, sizeof(tm_proc_items));
	int i;
	for(i=0; items[i].name; i++)
		items[i].data = selected;
	TWidget *wid = ttk_new_menu_widget(items, ttk_menufont, ret->w, ret->h);
	if(!wid) { pz_error(_("Unable to create TWidget!")); goto E1; }
	ttk_add_widget(ret, wid);
	
	// Add invisible widget for timer events
	wid = ttk_new_widget(0,0);
	if(!wid) { pz_error(_("Unable to create TWidget!")); goto E1; }
	wid->data = item;
	wid->frame = tm_selected_updater_frame;
	ttk_add_widget(ret, wid);
	ttk_widget_set_inv_fps(wid, 2);
	return pz_finish_window(ret);

E1:	ttk_free_window(ret);
E0:	return TTK_MENU_DONOTHING;
}

static PROC_ENT *tm_new_proc_ent(PROC_ENT *next) {
	PROC_ENT *ret;
	if(!(ret = calloc(1, sizeof(PROC_ENT)))) goto E0;
	if(!(ret->item = calloc(1, sizeof(ttk_menu_item)))) goto E1;
	ret->next = next;
	ret->pid = 0;
	ret->comm = NULL;
	*(ret->item) = (ttk_menu_item) { ret->comm, {tm_proc_mh}, TTK_MENU_ICON_SUB, ret, .free_name = 1, .free_data = 0 };
	return ret;
E1:	free(ret);
E0:	return NULL;
}

static void tm_free_proc_ent(PROC_ENT *ent) {
	if(!ent) return;
	if(ent->comm)
		free(ent->comm);
	free(ent);
}

// returns -1 if error, 1 if ent changed, 2 if ent's name changed, 0 otherwise
static int tm_update_proc_ent(PROC_ENT *ent) {
	static int page_size = 0;
	if(!page_size)
		page_size = sysconf(_SC_PAGESIZE);
	char path[40];
	char comm[20];
	PROC_ENT tmp = *ent;
	snprintf(path, sizeof(path), "/proc/%d/stat", ent->pid);
	FILE *file = fopen(path, "r");
	if(!file) goto E0;
	errno = 0;
	if(fscanf(file, "%*d (%19s %c %*d %*d %*d %*d %*d %*u %*u %*u %*u %*u %lu %lu"
		" %*d %*d %*d %ld %*d %*d %*u %lu",
		comm, &ent->state, &ent->utime, &ent->ktime, &ent->nice, &ent->vsize)
		== EOF && errno)
		goto E0;
	fclose(file);
	char *bracket = strrchr(comm, ')');
	if(bracket)
		(*bracket) = '\0';
	{
		int new_cent_stamp;
		int time_diff, udiff, kdiff;
		struct timeval tv;
		gettimeofday( &tv, 0 );
		new_cent_stamp = tv.tv_sec * 100 + tv.tv_usec / 10000;
		time_diff = new_cent_stamp - ent->cent_stamp;
		udiff = ent->utime - tmp.utime;
		kdiff = ent->ktime - tmp.ktime;
		if ((time_diff > 0) && (udiff >= 0) && (kdiff >= 0) ) {
			ent->uload = ((udiff << 22) / (time_diff << 6));
			ent->kload = ((kdiff << 22) / (time_diff << 6));
			if (ent->uload > 65536)
				ent->uload = 0;
			if (ent->kload > 65536)
				ent->kload = 0;
		} else {
			ent->kload = ent->uload = 0;
		}
		ent->cent_stamp = new_cent_stamp;
	}
	if(!ent->comm || strcmp(ent->comm, comm)) {
		if(ent->comm) free(ent->comm);
		ent->comm = strdup(comm);
		ent->item->name = ent->comm;
		return 1;
	}
	return 0;

E0:	fclose(file);
	return -1;
}

static void *tm_root_updater_md(TWindow *wwid) {
	UPDATER_DATA *ret = malloc(sizeof(UPDATER_DATA));
	if(!ret) goto E0;
	ret->wwid = wwid;
	if(!(ret->after = tm_new_proc_ent(NULL))) goto E1;
	if(!(ret->before = tm_new_proc_ent(ret->after))) goto E2;
	return ret;
E2:	free(ret->after);
E1:	free(ret);
E0:	return NULL;
}

static void tm_root_updater_swapentdata(PROC_ENT *a, PROC_ENT *b) {
	pid_t oldpid = a->pid;
	ttk_menu_item *olditem = a->item;
	char *oldcomm = a->comm;
	a->pid = b->pid;
	a->item = b->item;
	a->item->data = a;
	a->comm = b->comm;
	b->pid = oldpid;
	b->item = olditem;
	b->item->data = b;
	b->comm = oldcomm;
}

static void tm_root_updater_rment(PROC_ENT *cur, UPDATER_DATA *data) {
	PROC_ENT *next = cur->next;
	if(data->before->next == cur && next == data->after) {
		data->before->next = cur = data->after;
	} else {
		cur->next = next->next;
		tm_root_updater_swapentdata(cur, next);
		if(cur->next == NULL) data->after = cur;
	}
	tm_free_proc_ent(next);
}

static int tm_root_updater_frame(TWidget *wid) {
	DIR *proc = opendir("/proc");
	if(!proc) { pz_error(_("Unable to read from /proc!")); goto E0; }
	struct dirent *entry;
	
	UPDATER_DATA *data = wid->data;
	TWindow *wwid = data->wwid;
	TWidget *menu = wwid->focus;
	PROC_ENT *cur = data->before;

	while((entry = readdir(proc))) {
		if(!isdigit(entry->d_name[0])) continue;
		// next
		cur = cur->next;
		pid_t pid = atoi(entry->d_name);
		//if(pid >= 1000) break;
		while((cur != data->after) && (pid > cur->pid)) {
			ttk_menu_remove_by_ptr(menu, cur->item);
			tm_root_updater_rment(cur, data);
		}
		if(cur == data->after || pid < cur->pid) {
			PROC_ENT *new = tm_new_proc_ent(cur->next);
			if(!new) { pz_error(_("Could not allocate process entry!")); goto E1; }
			cur->next = new;
			tm_root_updater_swapentdata(cur, new);
			if(cur == data->after) {
				data->after = new;
				new->next = NULL;
			}
			cur->pid = pid;
			tm_update_proc_ent(cur);
			ttk_menu_append(menu, cur->item);
		} else {
			switch(tm_update_proc_ent(cur)) {
			case 1:
			case 2:
				ttk_menu_item_updated(menu, cur->item);
				break;
			case -1:
				ttk_menu_remove_by_ptr(menu, cur->item);
				tm_root_updater_rment(cur, data);
			}
		}
	}
	for(cur = cur->next; cur && cur != data->after; cur = cur->next) {
		ttk_menu_remove_by_ptr(menu, cur->item);
		tm_root_updater_rment(cur, data);
	}
	closedir(proc);
	
	// TODO: sort the menu with different comparators, eg, cpu usage, pid, comm
	ttk_menu_sort(menu);
	
	return 0;
	
E1:	closedir(proc);
E0:	pz_close_window(wwid);
	return 0;
}

static PzWindow *tm_root_mh(ttk_menu_item *item) {
	PzWindow *ret = pz_new_window(_("Task Manager"), PZ_WINDOW_NORMAL);
	if(!ret) { pz_error(_("Unable to create TWindow!")); goto E0;}
	//ttk_menu_item empty[1] = {NULL};
	TWidget *wid = ttk_new_menu_widget(NULL, ttk_menufont, ret->w, ret->h);
	if(!wid) { pz_error(_("Unable to create TWidget!")); goto E1; }
	ttk_add_widget(ret, wid);
	
	// Add invisible widget for timer events
	wid = ttk_new_widget(0,0);
	if(!wid) { pz_error(_("Unable to create TWidget!")); goto E1; }
	wid->data = tm_root_updater_md(ret);
	if(!wid->data) { pz_error(_("Unable to allocate memory for TWidget data!")); goto E2; }
	wid->frame = tm_root_updater_frame;
	ttk_widget_set_inv_fps(wid, 3);
	ttk_add_widget(ret, wid);
	// Populate menu
	tm_root_updater_frame(wid);
	return pz_finish_window(ret);
	
E2:	ttk_free_widget(wid);
E1:	ttk_free_window(ret);
E0:	return TTK_MENU_DONOTHING;
}

static void tm_cleanup() {
//	pz_free_config(tm_config);
// 	pz_menu_remove(_(TM_MENU_ROOT));
}

static void tm_init() {
	tm_module = pz_register_module("taskman", tm_cleanup);
	// TODO: Allow configurable display options
//	tm_config = pz_load_config(pz_module_get_cfgpath(module, "taskman"));
	
// 	PzConfItem *item
// 	if(!(item = pz_get_setting(tm_config, TM_VIEW_MODE)))
// 		item
	
	pz_menu_add_ttkh_group(_(TM_MENU_ROOT), _("System"), tm_root_mh, NULL);
}

PZ_MOD_INIT(tm_init)
